      *  FILECOPY

      *  demo/test program for IFS file copy with character set conversion
      *  uses no includes or copy members in order to be self sustained

      *  ILE RPG V5R1 (with V5R2 conditional compile)

      * please note that you can pass '0' for one of the CCSIDs, which means
      * "use job CCSID" (usually EBCDIC, unless one runs 65535)

      *  Christian Vitroler 19.06.2006

      *  no reason to run in a named activation group
     H DftActGrp(*NO) ActGrp(*CALLER)

      * calling quite a few UNIX style APIs
     H BndDir('QC2LE')


      *#########################################################################
      *  definitions
      *#########################################################################

      * constants --------------------------------------------------------------
     D True            C                   *ON
     D False           C                   *OFF

     D cMaxUnsigned    C                   4294967295

      * flags for open() and fcntl(); QSYSINC/H/FCNTL
     D O_RDONLY        C                   1
     D O_WRONLY        C                   2
     D O_RDWR          C                   4
     D O_CREAT         C                   8
     D O_EXCL          C                   16
     D O_TRUNC         C                   64
     D O_APPEND        C                   256

      * flags for chmode() and open(); QSYSINC/SYS/STAT
     D S_IRUSR         C                   256
     D S_IWUSR         C                   128
     D S_IXUSR         C                   64
     D S_IRWXU         C                   448

     D S_IRGRP         C                   32
     D S_IWGRP         C                   16
     D S_IXGRP         C                   8
     D S_IRWXG         C                   56

     D S_IROTH         C                   4
     D S_IWOTH         C                   2
     D S_IXOTH         C                   1
     D S_IRWXO         C                   7

      * constants for PathType in the TQlgPathName structure; QSYSINC/H/QLG
     D QLG_CHAR_SINGLE...
     D                 C                   0
     D QLG_PTR_SINGLE  C                   1
     D QLG_CHAR_DOUBLE...
     D                 C                   2
     D QLG_PTR_DOUBLE  C                   3

      * constants used by Qp0lGetAttr() and Qp0lSetAttr(); QSYSINC/H/QP0LSTDI
      * indicator for following a symlink
     D QP0L_DONOT_FOLLOW_SYMLNK...
     D                 C                   0
     D QP0L_FOLLOW_SYMLNK...
     D                 C                   1

      * attribute IDs
     D QP0L_ATTR_CCSID...
     D                 C                   27

      * UNIX error codes, more or less (IBM is a bit peculiar); QSYSINC/H/ERRNO
     D EINVAL          C                   3021
     D E2BIG           C                   3491


      * datatypes --------------------------------------------------------------
     D PDataType       S               *
     D TFilePath       S             64    Varying Based(PDataType)

      * NLS enabled path name structure
     D TQlgPathName    DS                  Qualified Based(PDataType)
     D   CCSID                       10I 0
     D   CountryID                    2
     D   LangID                       3
     D   Reserved1                    3
     D   PathType                    10U 0
     D   PathLen                     10I 0
     D   PathDelim                    2
     D   Reserved2                   10
     D   PathName                   256
     D   PathPtr                       *   Overlay(PathName)

      * describes CCSID plus options used for conversion from/to
     D TQtqCode        DS                  Qualified Based(PDataType)
     D   CCSID                       10I 0
     D   CnvAlter                    10I 0
     D   SubstAlter                  10I 0
     D   ShiftAlter                  10I 0
     D   LengthOption                10I 0
     D   MixedErrOpt                 10I 0
     D   Reserved                     8

      * conversion descriptor used by conversion APIs
     D TIconv          DS                  Qualified Based(PDataType)
     D   return_value                10I 0
     D   cd                          10I 0 Dim(12)

      * Qp0lGetAttr() and Qp0lSetAttr() attribute data header
     D TQp0lAttrHdr    DS                  Qualified Based(PDataType)
     D   NextAttrOffs                10U 0
     D   AttrID                      10U 0
     D   AttrSize                    10U 0
     D   Reserved                     4


      * program prototype ------------------------------------------------------
     D COPYFILE        PR                  ExtPgm('COPYFILE')
     D   iSrcPath                    64    Const
     D   iDstPath                    64    Const
     D   iSrcCCSID                   10I 0 Const
     D   iDstCCSID                   10I 0 Const


      * local prototypes
     D CopyFilePrim    PR            10I 0
     D   iSrcPath                          Value Like(TFilePath)
     D   iDstPath                          Value Like(TFilePath)
     D   iSrcCCSID                   10I 0 Const Options(*NoPass)
     D   iDstCCSID                   10I 0 Const Options(*NoPass)

     D CopyData        PR            10I 0
     D   ifhSrc                      10I 0 Const
     D   ifhDst                      10I 0 Const

     D ConvertData     PR            10I 0
     D   ifhSrc                      10I 0 Const
     D   ifhDst                      10I 0 Const
     D   iSrcCCSID                   10I 0 Const
     D   iDstCCSID                   10I 0 Const

     D SetFileCCSID    PR            10I 0
     D   iFilePath                         Value Like(TFilePath)
     D   iFileCCSID                  10I 0 Const

     D InitQlgPathPtr  PR
     D   aQlgPathName                      LikeDS(TQlgPathName)
     D   iPathPtr                      *   Value
     D   iPathSize                   10I 0 Value

     D ErrNo           PR            10I 0


      * API prototypes
      * http://publib.boulder.ibm.com/infocenter/iseries/v5r4/topic/apis/unix2.htm
     D Open            PR            10I 0 ExtProc('open')
     D  iFileName                      *   Value Options(*String)
     D  iOpenFlags                   10I 0 Value
     D  iMode                        10U 0 Value Options(*NoPass)
     D  iCCSID                       10U 0 Value Options(*NoPass)
      /if defined(*V5R2M0)
     D  iCrtCCSID                    10U 0 Value Options(*NoPass)
      /endif

     D CloseF          PR            10I 0 ExtProc('close')
     D  iFileHandle                  10I 0 Value

     D Read            PR            10I 0 ExtProc('read')
     D  iFileHandle                  10I 0 Value
     D  iBufferPtr                     *   Value
     D  iBufferSize                  10U 0 Value

     D Write           PR            10I 0 ExtProc('write')
     D  iFileHandle                  10I 0 Value
     D  iBufferPtr                     *   Value
     D  iBufferSize                  10U 0 Value

     D Qp0lSetAttr     PR            10I 0 ExtProc('Qp0lSetAttr')
     D   iQlgPathName                      Like(TQlgPathName)
     D   iBuffer                     26    Options(*VarSize)
     D   iBufferSize                 10U 0 Value
     D   iFlwSymLnk                  10U 0 Value

      * http://publib.boulder.ibm.com/infocenter/iseries/v5r4/topic/apis/nls3.htm
     D QtqIconvOpen    PR                  LikeDS(TIconv)
     D                                     ExtProc('QtqIconvOpen')
     D   iToCode                           Const LikeDS(TQtqCode)
     D   iFromCode                         Const LikeDS(TQtqCode)

     D iconv_close     PR            10I 0 ExtProc('iconv_close')
     D   iCD                               Value LikeDS(TIconv)

     D iconv           PR            10U 0 ExtProc('iconv')
     D   iCD                               Value LikeDS(TIconv)
     D   iPtrInpBuff                   *   Const
     D   InpBytesLeft                10U 0
     D   iPtrOutBuff                   *   Const
     D   OutBytesLeft                10U 0

      *  move iCount bytes from memory address to memory address (ILE C/C++)
     D MemMove         PR                  ExtProc('memmove')
     D   iBuffDst                      *   Value
     D   iBuffSrc                      *   Value
     D   iCount                      10U 0 Value

      * http://publib.boulder.ibm.com/infocenter/iseries/v5r4/topic/apis/mh1.htm
      * send program message
     D QMHSNDPM        PR                  ExtPgm('QMHSNDPM')
     D   iMsgID                       7    Const
     D   iQualMsgFile                20    Const
     D   iMsgDataText             32767    Const Options(*VarSize)
     D   iMsgDataLen                 10I 0 Const
     D   iMsgType                    10    Const
     D   iCallStkEnt                 10    Const Options(*VarSize)
     D   iCallStkCnt                 10I 0 Const
     D   oMsgKey                      4
     D   aError                   32767    Options(*VarSize)
      * first optional group
     D   iCallStkLen                 10I 0 Const Options(*NoPass)
     D   iCallStkQual                20    Const Options(*NoPass)
     D   iDspMsgWait                 10I 0 Const Options(*NoPass)
      * second optional group
     D   iCallStkType                10    Const Options(*NoPass)
     D   iDataCCSID                  10I 0 Const Options(*NoPass)


      * global variables -------------------------------------------------------
     D rc              S             10I 0
     D msgText         S             78
     D msgLen          S             10I 0
     D msgKey          S              4

     D msgError        DS                  Qualified
     D   NbrBytesPrv                 10I 0 Inz(0)
     D   NbrBytesAvl                 10I 0 Inz(0)


      *#########################################################################
      * program interface definition & global keys
      *#########################################################################
     D COPYFILE        PI
     D   iSrcPath                    64    Const
     D   iDstPath                    64    Const
     D   iSrcCCSID                   10I 0 Const
     D   iDstCCSID                   10I 0 Const


      *#########################################################################
      *  main program
      *#########################################################################
      /free

       if (iSrcCCSID = 0) and (iDstCCSID = 0);
         rc = CopyFilePrim(%Trim(iSrcPath): %Trim(iDstPath));
       else;
         rc = CopyFilePrim
                (%Trim(iSrcPath): %Trim(iDstPath): iSrcCCSID: iDstCCSID);
       endif;

       if (rc <> 0);
         msgText = 'error during ' + %trimr(msgText) + ' (' + %char(rc) + ')';
         msgLen = %len(%trimr(msgText));
         QMHSNDPM('CPF9897': 'QCPFMSG   *LIBL': msgText: msgLen: '*DIAG':
            '*PGMBDY': 1: msgKey: msgError);
       endif;

       *INLR = True;
       return;

      /end-free


      *#########################################################################
      *  application specific procedures
      *#########################################################################


      ***CopyFilePrim***********************************************************
      *  copy a file within the IFS with optional character set translation
      *-------------------------------------------------------------------------
      *  While it is possible to open a file with the option to create the file
      *  if it does not exist or truncate it in case it does, this concept has
      *  a serious design flaw when it comes to the file's CCSID. The very same
      *  call will set the CCSID properly in the case of file creation, but will
      *  completely ignore it during a truncation.
      *
      *  Now, we could simply check first if the file exists and in that case
      *  change the CCSID & truncate it on open and open it with a CCSID in the
      *  other case. That'll add another API call, i.e. "access()", and some
      *  unnecessary complex if/else
      *
      *  I'll take another approach. Won't care about CCSID upon creation, which
      *  does not matter as we do the conversion ourselves, and simply change
      *  the CCSID afterwards. Who cares if it existed already or we created it?
      **************************************************************************
     P CopyFilePrim    B
     D CopyFilePrim    PI            10I 0
     D   iSrcPath                          Value Like(TFilePath)
     D   iDstPath                          Value Like(TFilePath)
     D   iSrcCCSID                   10I 0 Const Options(*NoPass)
     D   iDstCCSID                   10I 0 Const Options(*NoPass)

     D result          S             10I 0
     D needConv        S               N
     D fhSrc           S             10I 0
     D fhDst           S             10I 0
     D options         S             10I 0
     D mode            S             10I 0
      *-------------------------------------------------------------------------
      /free

       // it takes two (CCSIDs) to do a proper conversion
       msgText = 'checking params';
       if (%Parms = 3);
         return -1;
       endif;
       needConv = (%Parms >= 4) and (iSrcCCSID <> iDstCCSID);

       // open the source file in "raw" mode as we do the conversion ourselves
       msgText = 'opening source file';
       fhSrc = Open(iSrcPath: O_RDONLY);
       if (fhSrc < 0);
         return ErrNo();
       endif;

       // open the dest. file in "raw" mode as we do the conversion ourselves
       msgText = 'opening destination file';
       options = O_WRONLY + O_CREAT + O_TRUNC;
       mode = S_IRWXU + S_IRWXG + S_IRWXO;

       fhDst = Open(iDstPath: options: mode);
       if (fhDst < 0);
         result = ErrNo();
         CloseF(fhSrc);
         return result;
       endif;

       // do the actual conversion or plain copy
       if NeedConv;
         result = ConvertData(fhSrc: fhDst: iSrcCCSID: iDstCCSID);
       else;
         result = CopyData(fhSrc: fhDst);
       endif;

       // close the files
       CloseF(fhSrc);
       CloseF(fhDst);

       if (result = 0) and NeedConv;
         msgText = 'setting CCSID on out file';
         result = SetFileCCSID(iDstPath: iDstCCSID);
       endif;

       return result;

      /end-free
     P                 E


      ***CopyData***************************************************************
      *  copy data from one file to the other in a simple loop
      **************************************************************************
     P CopyData        B
     D CopyData        PI            10I 0
     D   ifhSrc                      10I 0 Const
     D   ifhDst                      10I 0 Const

     D result          S             10I 0
     D nRead           S             10I 0
     D nWritten        S             10I 0

     D buffer          S          65532
      *-------------------------------------------------------------------------
      /free

       msgText = 'file copy';
       nRead = Read(ifhSrc: %Addr(buffer): %Size(buffer));
       dow (nRead > 0);
         if (Write(ifhDst: %Addr(buffer): nRead) < 0);
           return -1;
         endif;
         nRead = Read(ifhSrc: %Addr(buffer): %Size(buffer));
       enddo;

       return nRead;

      /end-free
     P                 E


      ***ConvertData************************************************************
      *  copy data from one file to the other with character conversion
      *-------------------------------------------------------------------------
      *  supports SBCS, DBCS and Mixed character sets, including UTF-8
      *  takes care of the possible conraction or expansion of data as well as
      *  character truncation upon read from file
      *  note: if the output buffer would be smaller than the largest character
      *        we'll loop forever (e.g. 2 bytes & UTF-8)
      *  note: if using a 3 byte output buffer and input data is UTF-8 and
      *        contains a "zero width no-break space" as byte order marker, then
      *        iconv() throws an EUNKNOWN (API is calling for help; IBM systems
      *        engineer needed)
      **************************************************************************
     P ConvertData     B
     D ConvertData     PI            10I 0
     D   ifhSrc                      10I 0 Const
     D   ifhDst                      10I 0 Const
     D   iSrcCCSID                   10I 0 Const
     D   iDstCCSID                   10I 0 Const

     D result          S             10I 0
     D nRead           S             10I 0
     D nWritten        S             10I 0

     D inpBuffer       S          65532
     D outBuffer       S          65532

     D inpPtr          S               *
     D outPtr          S               *

     D inpLen          S             10U 0
     D outLen          S             10U 0

     D srcQtQCode      DS                  LikeDS(TQtQCode)
     D dstQtQCode      DS                  LikeDS(TQtQCode)
     D cd              DS                  LikeDS(TIconv)
     D rcu             S             10U 0
      *-------------------------------------------------------------------------
      /free

       // let's get a conversion descriptor
       msgText = 'creating conversion descriptor file';
       srcQtQCode = *Allx'00';
       srcQtQCode.CCSID = iSrcCCSID;
       dstQtQCode = *Allx'00';
       dstQtQCode.CCSID = iDstCCSID;

       cd = QtqIconvOpen(dstQtQCode: srcQtQCode);
       if (cd.return_value < 0);
         result = ErrNo();
         return result;
       endif;

       msgText = 'file conversion';
       inpLen = 0;
       outLen = %Size(outBuffer);
       outPtr = %Addr(outBuffer);

       dow (outLen > 0);
         inpPtr = %Addr(inpBuffer);
         nRead = Read(ifhSrc: inpPtr + inpLen: %Size(inpBuffer) - inpLen);

         select;
         when (nRead = 0) and (inpLen = 0);
           // file is done, conversion is done; let's flush what we have left
           exsr FlushOutBuffer;
           leave;

         when (nread < 0);
           // error while reading from file;
           result = ErrNo;
           leave;
         endsl;

         inpLen = inpLen + nRead;
         rcu = iconv(cd: inpPtr: inpLen: outPtr: outLen);

         // some less than usual condition occured
         if (rcu = cMaxUnsigned);
           select;
           when (ErrNo = EINVAL);
             // there is a partial character at the end of the input buffer
             if (nRead = 0);
               // too bad, we got an orphan; anyway, let's write what we have
               exsr FlushOutBuffer;
               // end then leave in shame (?)
               // result = -1;
               leave;
             else;
               // let's carry it over into the input buffer
               MemMove(%Addr(inpBuffer): inpPtr: inpLen);
             endif;

           when (ErrNo = E2BIG);
             // output buffer is full while still some chars in input buffer
             // let's flush the output buffer
             exsr FlushOutBuffer;
             // let's carry the remaining chars over into the input buffer
             MemMove(%Addr(inpBuffer): inpPtr: inpLen);

           other;
             // Houston, we've got a problem!
             result = ErrNo();
             leave;
           endsl;
         else;

           if (outLen = 0);
             // out buffer is full, flush it
             // happens all the time if converting from SBCS to SBCS
             ExSR FlushOutBuffer;
           endif;

         endif;  // if (rcu = cMaxUnsigned);
       enddo;

       iconv_close(cd);
       return result;


       // ----------------------------------------------------------------------
       begsr FlushOutBuffer;
         if (Write(ifhDst: %Addr(outBuffer): %Size(outBuffer) - outLen) >= 0);
           outPtr = %Addr(outBuffer);
           outLen = %Size(outBuffer);
         else;
           result = ErrNo;
           outLen = 0;  // ends the loop
         endif;
       endsr;

      /end-free
     P                 E


      ***SetFileCCSID***********************************************************
      *  set the CCSID of a file in the IFS using Qp0lSetAttr
      *-------------------------------------------------------------------------
      *  http://publib.boulder.ibm.com/infocenter/iseries/v5r4/topic/apis/
      *         qsetattr.htm
      **************************************************************************
     P SetFileCCSID    B
     D SetFileCCSID    PI            10I 0
     D   iFilePath                         Value Like(TFilePath)
     D   iFileCCSID                  10I 0 Const

     D QlgPathName     DS                  LikeDS(TQlgPathName)

      /if defined(*V5R2M0)
     D AttrBuffer      DS                  Qualified
     D   Header                            LikeDS(TQp0lAttrHdr)
     D   CCSID                       10U 0

      /else
     D AttrBuffer      DS                  Qualified
     D   Header                            Like(TQp0lAttrHdr)
     D   CCSID                       10U 0

     D PtrHeader       S               *   Inz(%Addr(AttrBuffer.Header))
     D Header          DS                  LikeDS(TQp0lAttrHdr) Based(PtrHeader)
     D
      /endif
      *-------------------------------------------------------------------------
      /free

       InitQlgPathPtr(QlgPathName: %Addr(iFilePath) + 2: %Len(iFilePath));

      /if defined(*V5R2M0)
       AttrBuffer.Header = *Allx'00';
       AttrBuffer.Header.AttrID = QP0L_ATTR_CCSID;
       AttrBuffer.Header.AttrSize = %Size(AttrBuffer.CCSID);

      /else
       Header = *Allx'00';
       Header.AttrID = QP0L_ATTR_CCSID;
       Header.AttrSize = %Size(AttrBuffer.CCSID);
      /endif

       AttrBuffer.CCSID = iFileCCSID;

       Return Qp0lSetAttr
         (QlgPathName: AttrBuffer: %Size(AttrBuffer): QP0L_FOLLOW_SYMLNK);

      /end-free
     P                 E


      ***InitQlgPathPtr*********************************************************
      *  initialize a TQlgPathName using a pointer to the path name
      *-------------------------------------------------------------------------
      *  this procedure assumes that we are using a single byte slash as path
      *  seperator and the job's CCSID, country & language settings
      *  note: using a pointer eliminates any restriction on the length of the
      *        path name
      **************************************************************************
     P InitQlgPathPtr  B                   Export
     D InitQlgPathPtr  PI
     D   aQlgPathName                      LikeDS(TQlgPathName)
     D   iPathPtr                      *   Value
     D   iPathSize                   10I 0 Value
      *-------------------------------------------------------------------------
      /free

       aQlgPathName = *Allx'00';
       aQlgPathName.PathDelim = '/';
       aQlgPathName.PathType = QLG_PTR_SINGLE;
       aQlgPathName.PathPtr = iPathPtr;
       aQlgPathName.PathLen = iPathSize;

      /end-free
     P InitQlgPathPtr  E


      ***ErrNo******************************************************************
      *  return error number for Unix/C APIs
      *-------------------------------------------------------------------------
      *  most, if not all UNIX style APIs, and also quite a few other C APIs
      *  return only failure; the exact error condition is to be found in the
      *  global variable errno
      *  we have to call the external procedure __errno, which gets us the
      *  POINTER to the actual errno (which is not really that global, but a
      *  thread safe variable)
      *  if you treat errno as a hex number and add the prefix 'CPE', you get
      *  a proper i5/OS message ID, in case you need one
      *
      *  reference: QSYSINC/H/ERRNO
      **************************************************************************
     P ErrNo           B
     D ErrNo           PI            10I 0

     D GetPtrErrNo     PR              *   ExtProc('__errno')

     D PErrNo          S               *
     D result          S             10I 0 Based(PErrNo)
      *-------------------------------------------------------------------------
      /free

       PErrNo = GetPtrErrNo;
       return result;

      /end-free
     P                 E

